from sanic import views, response
from merkava.channels import Channel


class BaseChannelView(views.HTTPMethodView):

    def set_channel(self, request, channel_name):
        self.channel = Channel(
            channel_name=channel_name,
            data_path=request.app.config.get('path', './'),
        )

    # service.add_task(self.channel.listen())

    def dispatch_request(self, request, *args, **kwargs):
        self.set_channel(request, kwargs.get('channel'))
        return super().dispatch_request(request, *args, **kwargs)


class ChannelView(BaseChannelView):

    async def post(self, request, channel, id=None):
        # TODO:
        # - make the CRUD operations async/await
        status, result = self.channel.create(request.json)
        return response.json({'result': result}, status=status)

    async def get(self, request, channel, id):
        # TODO:
        # - make the CRUD operations async/await
        status, result = self.channel.retrieve(id)
        return response.json(result, status=status)

    async def patch(self, request, channel, id):
        # TODO:
        # - make the CRUD operations async/await
        status, result = self.channel.update(id, request.json)
        return response.json(result, status=status)

    async def delete(self, request, channel, id):
        # TODO:
        # - make the CRUD operations async/await
        status, result = self.channel.delete(id)
        return response.json(None, status=status)

    async def put(self, request, channel, id):
        # TODO:
        # - make the CRUD operations async/await
        status, result = self.channel.restore(id)
        return response.json(result, status=status)


class RecentChannelView(BaseChannelView):

    async def get(self, request, channel, num=None):
        if not num:
            num = 5
        num = int(num)
        result = self.channel.recent(int(num))
        return response.json(result)



import configparser
import sys
import os

from merkava import blueprints
from sanic import Sanic
from sanic.response import json

if os.environ.get('DEVELOPMENT') == 'true':  # pragma: no cover
    from aoiklivereload import LiveReloader

    reloader = LiveReloader()
    reloader.start_watcher_thread()
service = Sanic()
service.blueprint(blueprints.bp)


@service.route("/")
async def is_alive(request):
    return json({'is_alive': True})


@service.listener('before_server_start')
async def get_config(app, loop):
    if len(sys.argv) > 1:  # pragma: no cover
        path = sys.argv[1]
        if os.path.isfile(path):
            config = configparser.ConfigParser()
            config.read(path)
            app.config.update(config['Storage'])


if __name__ == '__main__':  # pragma: no cover
    service.run(host="127.0.0.1", port=6363)



from sanic import Blueprint
from merkava import views

bp = Blueprint('channels', url_prefix='/v1/<channel>')
bp.add_route(views.ChannelView.as_view(), '/<id:[0-9]*>')
bp.add_route(views.RecentChannelView.as_view(), '/recent/<num:[0-9]*>')



import merkava
import os
import json

service = merkava.server.service
path = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'data')
channel_name = 'text_channel'


class TestMerkava(object):

    @classmethod
    def setup_class(cls):
        service.config.update({'path': path})

    def url(self, *args):
        url = os.path.join(
            '/', 'v1', 'text_channel', *(str(arg) for arg in args)
        )
        if not url.endswith('/'):
            url += '/'
        return url

    ######################################
    #   Basic CRUD and operation tests   #
    ######################################

    def test_channel_create_index(self):
        _, response = service.test_client.get(self.url('9999'))
        file_path = os.path.join(path, '{}.mrkv'.format(channel_name))
        assert os.path.exists(file_path)
        assert os.path.isfile(file_path)

    def test_retrieve_not_exists(self):
        _, response = service.test_client.get(self.url('9999'))
        assert response.json is None
        assert response.status == 404

    def test_create(self):
        data = {'foo': 'bar', 'action': 'create'}
        _, response = service.test_client.post(
            self.url(), data=json.dumps(data)
        )
        assert response.json is not None
        assert response.status == 201
        assert response.json.get('result').get('data') == data
        assert response.json.get('result').get('data').get(
            'action'
        ) == 'create'

    def test_retrieve(self):
        data = {'foo': 'bar', 'action': 'retrieve'}
        service.test_client.post(self.url(), data=json.dumps(data))
        _, response = service.test_client.get(self.url(2))
        assert response.json is not None
        assert response.status == 200
        assert response.json.get('data') == data
        assert response.json.get('data').get('action') == 'retrieve'
        assert response.json.get('id') == 2

    def test_update(self):
        data = {'one': 1, 'two': 2}
        update = {'two': 200, 'three': 3}
        service.test_client.post(self.url(), data=json.dumps(data))
        _, response = service.test_client.patch(
            self.url(3), data=json.dumps(update)
        )
        assert response.status == 200
        _, response = service.test_client.get(self.url(3))
        assert response.json is not None
        assert response.json.get('data') != data
        assert response.json.get('id') == 3
        assert response.json.get('data').get('one') == 1
        assert 'two' in response.json.get('data')
        assert 'three' in response.json.get('data')
        assert response.json.get('data').get('two') == 200
        assert response.json.get('data').get('three') == 3

    def test_delete(self):
        data = {'foo': 'bar', 'action': 'delete'}
        service.test_client.post(self.url(), data=json.dumps(data))
        _, response = service.test_client.delete(self.url(4))
        assert response.status == 204
        assert response.json is None
        _, response = service.test_client.get(self.url(4))
        assert response.json is None
        assert response.status == 404

    def test_restore(self):
        data = {'foo': 'bar', 'action': 'restore'}
        service.test_client.post(self.url(), data=json.dumps(data))
        service.test_client.delete(self.url(5))
        _, response = service.test_client.put(self.url(5))
        assert response.status == 200
        assert response.json is not None
        assert response.json.get('data') == data
        assert response.json.get('id') == 5
        assert response.json.get('data').get('action') == 'restore'
        assert response.json.get('is_deleted') is False
        assert response.json.get('was_restored') is True

    def test_recent(self):
        _, response = service.test_client.get(self.url('recent'))
        assert response.json is not None
        assert response.status == 200
        assert isinstance(response.json, list)
        assert len(response.json) > 0
        _, response = service.test_client.get(self.url('recent', 2))
        assert response.json is not None
        assert response.status == 200
        assert isinstance(response.json, list)
        assert len(response.json) == 2

    ###########################################
    #   Specific feature and use case tests   #
    ###########################################

    def test_storage_of_string(self):
        data = 'foo'
        service.test_client.post(self.url(), data=json.dumps(data))
        _, response = service.test_client.get(self.url(6))
        assert response.json is not None
        assert response.status == 200
        assert response.json.get('data') == data
        assert isinstance(response.json.get('data'), str)

    def test_storage_of_boolean(self):
        data = False
        service.test_client.post(self.url(), data=json.dumps(data))
        _, response = service.test_client.get(self.url(7))
        assert response.json is not None
        assert response.status == 200
        assert response.json.get('data') == data
        assert isinstance(response.json.get('data'), bool)
        data = True
        service.test_client.post(self.url(), data=json.dumps(data))
        _, response = service.test_client.get(self.url(8))
        assert response.json is not None
        assert response.status == 200
        assert response.json.get('data') == data
        assert isinstance(response.json.get('data'), bool)

    def test_storage_of_none(self):
        data = None
        service.test_client.post(self.url(), data=json.dumps(data))
        _, response = service.test_client.get(self.url(9))
        assert response.json is not None
        assert response.status == 200
        assert response.json.get('data') == data
        assert response.json.get('data') is None

    def test_storage_of_int(self):
        data = 1
        service.test_client.post(self.url(), data=json.dumps(data))
        _, response = service.test_client.get(self.url(10))
        assert response.json is not None
        assert response.status == 200
        assert response.json.get('data') == data
        assert isinstance(response.json.get('data'), int)

    def test_storage_of_float(self):
        data = 1.0
        service.test_client.post(self.url(), data=json.dumps(data))
        _, response = service.test_client.get(self.url(11))
        assert response.json is not None
        assert response.status == 200
        assert response.json.get('data') == data
        assert isinstance(response.json.get('data'), float)

    def test_storage_of_list(self):
        data = [65535, (65536 * 65536), 0x7fffffff, 0x80000000, 1e10]
        service.test_client.post(self.url(), data=json.dumps(data))
        _, response = service.test_client.get(self.url(12))
        assert response.json is not None
        assert response.status == 200
        assert response.json.get('data') == data
        assert all(
            isinstance(x, (float, int)) for x in response.json.get('data')
        )

    def test_update_of_non_dict(self):
        data = 'foo'
        update = 'bar'
        service.test_client.post(self.url(), data=json.dumps(data))
        _, response = service.test_client.patch(
            self.url(13), data=json.dumps(update)
        )
        assert response.status == 200
        _, response = service.test_client.get(self.url(13))
        assert response.json is not None
        assert response.json.get('data') != data
        assert response.json.get('id') == 13
        assert response.json.get('data') == update

    @classmethod
    def teardown_class(cls):
        folder = service.config.get('path')
        if os.path.exists(folder):
            for file in os.listdir(folder):
                file_path = os.path.join(folder, file)
                if os.path.isfile(file_path) and '.gitkeep' not in file_path:
                    os.unlink(file_path)



import os
import msgpack
from datetime import datetime
import pytz


# import asyncio


class Result:

    def __init__(self, data):
        self.data = data

    @property
    def is_deleted(self):
        return self.data.get(b'is_deleted', False)


class Channel:
    # TODO:
    # - Set maximum_recent to configurable value
    maximum_recent = 25

    def __init__(self, channel_name, data_path):
        self.channel_name = channel_name.lower()
        self.data_path = data_path
        self.index_path = os.path.join(
            data_path, '{}.mrkv'.format(self.channel_name)
        )
        if not os.path.exists(self.index_path):
            self.index = 1
            self._store_index()
        else:
            self.index = self._load(self.index_path)

    # async def listen(self):
    #     with open(self.index_path, 'rb') as file:
    #         file.seek(0, 0)
    #         while True:
    #             line = file.readline()
    #             if not line:
    #                 await asyncio.sleep(0.1)
    #                 continue
    #             print(line)
    #             yield line

    def create(self, raw):
        file_path = self._build_path(self.index)
        index = self.index
        data = {
            'id': index,
            'created': datetime.now(tz=pytz.utc).isoformat(),
            'data': raw,
        }
        self._dump(file_path, data)
        self += 1
        return 201, data

    def retrieve(self, id, override=False):
        file_path = self._build_path(id)
        if not os.path.exists(file_path):
            return 404, None

        data = self._load(file_path)
        if data.get(b'is_deleted', False) and not override:
            return 404, None

        return 200, data

    def update(self, id, raw):
        file_path = self._build_path(id)
        _, data = self.retrieve(id)
        inner_data = data.get(b'data')
        if isinstance(inner_data, dict):
            data.get(b'data').update(raw)
        else:
            data[b'data'] = raw
        self._dump(file_path, data)
        return 200, data

    def delete(self, id):
        file_path = self._build_path(id)
        _, data = self.retrieve(id)
        data.update({'is_deleted': True})
        self._dump(file_path, data)
        return 204, {'ok': True}

    def restore(self, id):
        file_path = self._build_path(id)
        _, data = self.retrieve(id, override=True)
        data.update({'is_deleted': False, 'was_restored': True})
        self._dump(file_path, data)
        return 200, data

    def recent(self, num):
        results = []
        index = self.index - 1
        num = min(num, self.maximum_recent)
        while len(results) < num and index > 0:
            result = self.get_result(index)
            if result is not None and not result.is_deleted:
                results.append(result.data)
            index -= 1
        return results

    def get_result(self, id):
        _, raw = self.retrieve(id)
        if raw is None:
            return None

        return Result(raw)

    def __add__(self, n):
        self.index += n
        self._store_index()
        return self.index

    def _build_path(self, id):
        file_name = '{}.{}.mrkv'.format(self.channel_name, id)
        return os.path.join(self.data_path, file_name)

    def _load(self, path):
        with open(path, 'rb') as store:
            return msgpack.load(store, use_list=False)

    def _dump(self, path, data):
        with open(path, 'wb') as store:
            msgpack.dump(data, store)

    def _store_index(self):
        self._dump(self.index_path, self.index)
